section \<open>Semantics in labeled graphs\<close>

theory LabeledGraphSemantics
imports MissingRelation
begin

text \<open>The thing called term in the paper is called @{term alligorical_term} here.
      This naming is chosen because an allegory has precisely these operations, plus identity. \<close>

datatype 'l allegorical_term
 = A_Int "'l allegorical_term" "'l allegorical_term" (infix "\<sqinter>" 65)
 | A_Cmp "'l allegorical_term" "'l allegorical_term" (infix ";" 70)
 | A_Cnv "'l allegorical_term" ("_\<^sup>-" 80)
 | A_Atm 'l

text \<open>The interpretation of terms, Definition 2.\<close>
fun semantics :: "('l \<Rightarrow> ('v \<times> 'v) set) \<Rightarrow> 'l allegorical_term \<Rightarrow> ('v \<times> 'v) set" (":_:\<lbrakk>_\<rbrakk>" 55) where
":G:\<lbrakk>a \<sqinter> b\<rbrakk> = (:G:\<lbrakk>a\<rbrakk>) \<inter> (:G:\<lbrakk>b\<rbrakk>)" |
":G:\<lbrakk>a ; b\<rbrakk> = (:G:\<lbrakk>a\<rbrakk>) O (:G:\<lbrakk>b\<rbrakk>)" |
":G:\<lbrakk>a\<^sup>-\<rbrakk> = (:G:\<lbrakk>a\<rbrakk>)\<inverse>" |
":G:\<lbrakk>A_Atm l\<rbrakk> = G l"

type_synonym 'v sentence = "'v allegorical_term \<times> 'v allegorical_term"

text \<open>Definition 3. We don't define sentences but instead simply work with pairs of terms.\<close>
abbreviation holds :: "('l \<Rightarrow> ('v \<times> 'v) set) \<Rightarrow> 'l allegorical_term \<times> 'l allegorical_term \<Rightarrow> bool" (infix "\<Turnstile>" 55) where
"G \<Turnstile> S \<equiv> :G:\<lbrakk>fst S\<rbrakk> = :G:\<lbrakk>snd S\<rbrakk>"

abbreviation subset_sentence (infix "\<sqsubseteq>" 60) where
"A \<sqsubseteq> B \<equiv> (A,A_Int A B)"

text \<open>Lemma 1.\<close>
lemma sentence_iff[simp]:
  "G \<Turnstile> e\<^sub>1 \<sqsubseteq> e\<^sub>2 = (:G:\<lbrakk>e\<^sub>1\<rbrakk> \<subseteq> :G:\<lbrakk>e\<^sub>2\<rbrakk>)" and
  eq_as_subsets:
  "G \<Turnstile> (e\<^sub>1, e\<^sub>2) = (G \<Turnstile> e\<^sub>1 \<sqsubseteq> e\<^sub>2 \<and> G \<Turnstile> e\<^sub>2 \<sqsubseteq> e\<^sub>1)"
  by auto

lemma image_semantics_move_outw[intro]:
  shows "map_prod f f ` (:G:\<lbrakk>e\<rbrakk>) \<subseteq> :\<lambda> x. map_prod f f ` G x:\<lbrakk>e\<rbrakk>"
  by(induct e,auto)

lemma map_graph_in[intro]:
  assumes "(a,b) \<in> :G:\<lbrakk>e\<rbrakk>"
  shows "(f a,f b) \<in> :\<lambda> x. map_prod f f ` G x:\<lbrakk>e\<rbrakk>"
  using assms image_semantics_move_outw by blast

lemma semantics_final_step:
  assumes "(a,b) \<in> :A:\<lbrakk>e\<rbrakk>"
  shows "\<exists> b' x. (a,b') \<in> A x \<or> (b',a) \<in> A x"
        "\<exists> a' x. (b,a') \<in> A x \<or> (a',b) \<in> A x"
  using assms by(atomize(full),induct e arbitrary:a b,auto)

lemma graph_union_semantics:
  shows "(:A:\<lbrakk>e\<rbrakk>) \<union> (:B:\<lbrakk>e\<rbrakk>) \<subseteq> :(\<lambda> l. A l \<union> B l):\<lbrakk>e\<rbrakk>"
  by(induct e,auto)

lemma subset_semantics:
  assumes "\<And> l. A l \<subseteq> B l" "(a,b) \<in> :A:\<lbrakk>e\<rbrakk>"
  shows "(a,b) \<in> :B:\<lbrakk>e\<rbrakk>"
  using assms by(induct e arbitrary: a b,auto intro!:relcompI)
 

end