section \<open>Standard Models\<close>
text \<open>We define the kind of models we are interested in here.
      In particular, we care about standard graphs.
      To allow some reuse, we distinguish a generic version called @{term standard},
      from an instantiated abbreviation @{term standard'}.
      There is little we can prove about these definition here, except for Lemma 2.\<close>
theory StandardModels
imports LabeledGraphSemantics LabeledGraphs
begin


datatype ('a,'b) Standard_Constant = S_Bot | S_Top | S_Idt | S_Const 'a | S_Var 'b
abbreviation a_bot ("\<bottom>") where "\<bottom> \<equiv> A_Atm S_Bot" 
abbreviation a_top ("\<top>") where "\<top> \<equiv> A_Atm S_Top"
abbreviation a_idt ("\<one>") where "\<one> \<equiv> A_Atm S_Idt"

abbreviation ident_rel where
"ident_rel G idt \<equiv> getRel G idt = (\<lambda> x.(x,x)) ` vertices G"

lemma ident_relI[intro]:
  assumes min:"\<And> x. x \<in> vertices G \<Longrightarrow> (x,x) \<in> getRel G idt"
  and max1:"\<And> x y. (x,y) \<in> getRel G idt \<Longrightarrow> x = y"
  and max2:"\<And> x y. (x,y) \<in> getRel G idt \<Longrightarrow> x \<in> vertices G"
shows "ident_rel G idt"
proof
  from max1 max2 have "\<And> x y. (x,y) \<in> getRel G idt \<Longrightarrow> (x = y \<and> x \<in> vertices G)" by simp
  thus "getRel G idt \<subseteq> (\<lambda>x. (x, x)) ` vertices G" by auto
  show "(\<lambda>x. (x, x)) ` vertices G \<subseteq> getRel G idt" using min by auto
qed


text \<open>Definition 4, generically.\<close>
definition standard :: "('l \<times> 'v) set \<Rightarrow> 'l \<Rightarrow> 'l \<Rightarrow> 'l \<Rightarrow> ('l, 'v) labeled_graph \<Rightarrow> bool" where
"standard C b t idt G
   \<equiv> G = restrict G
   \<and> vertices G \<noteq> {}
   \<and> ident_rel G idt
   \<and> getRel G b = {}
   \<and> getRel G t = {(x,y). x\<in>vertices G \<and> y\<in>vertices G}
   \<and> (\<forall> (l,v) \<in> C. getRel G l = {(v,v)})"

type_synonym ('v,'l) std_term = "('v, 'l) Standard_Constant allegorical_term"
type_synonym ('v,'l) std_sentence = "('v, 'l) std_term \<times> ('v, 'l) std_term"
type_synonym ('c,'l,'v) std_graph = "(('c,'l) Standard_Constant, ('c+'v)) labeled_graph"

text \<open>Definition 4.\<close>
definition standard' :: "'v set \<Rightarrow> ('v,'l,'a) std_graph \<Rightarrow> bool" where
"standard' C G \<equiv> (\<forall> c. c \<in> C \<longleftrightarrow> Inl c \<in> vertices G) \<and>
  standard ((\<lambda> c. (S_Const c,Inl c)) ` C) S_Bot S_Top S_Idt G"

lemma standard'_def2:
"standard' C G = (graph G \<and>
     vertices G \<noteq> {} \<and>
     (\<forall> c. c \<in> C \<longleftrightarrow> Inl c \<in> vertices G) \<and>
     :getRel G:\<lbrakk>\<one>\<rbrakk> = (\<lambda>x. (x, x)) ` vertices G \<and>
     :getRel G:\<lbrakk>\<bottom>\<rbrakk> = {} \<and>
     :getRel G:\<lbrakk>\<top>\<rbrakk> = {(x, y). x \<in> vertices G \<and> y \<in> vertices G} \<and>
     (\<forall>c \<in> C. :getRel G:\<lbrakk>A_Atm (S_Const c)\<rbrakk> = {(Inl c, Inl c)}))" 
  unfolding standard_def standard'_def semantics.simps by auto

text \<open>Definition 5.\<close>
definition model :: "'v set \<Rightarrow> ('v,'l,'a) std_graph \<Rightarrow> ('v, 'l) std_sentence set \<Rightarrow> bool" where
"model C G T \<equiv> standard' C G \<and> (\<forall> S \<in> T. getRel G \<Turnstile> S)"

text \<open>Definition 5.\<close>
abbreviation consistent :: "'b itself \<Rightarrow> 'v set \<Rightarrow> (('v,'l) std_sentence) set \<Rightarrow> bool" where
"consistent _ C T \<equiv> \<exists> (G::('v,'l,'b) std_graph). model C G T"

text \<open>Definition 6.\<close>
definition entails :: "'b itself \<Rightarrow> 'v set \<Rightarrow> (('v,'l) std_sentence) set \<Rightarrow> ('v,'l) std_sentence \<Rightarrow> bool"  where
"entails _ C T S \<equiv> \<forall> (G::('v,'l,'b) std_graph). model C G T \<longrightarrow> getRel G \<Turnstile> S"

lemma standard_top_not_bot[intro]:
"standard' C G \<Longrightarrow> :getRel G:\<lbrakk>\<bottom>\<rbrakk> \<noteq> :getRel G:\<lbrakk>\<top>\<rbrakk>"
  unfolding standard_def standard'_def by auto

text \<open>Lemma 2.\<close>
lemma consistent_iff_entails_nonsense:
"consistent t C T = (\<not> entails t C T (\<bottom>,\<top>))"
proof
  show "consistent t C T \<Longrightarrow> \<not> entails t C T (\<bottom>, \<top>)"
    using standard_top_not_bot semantics.simps(4)
    unfolding entails_def model_def
    by (smt fst_conv snd_conv standard_top_not_bot)
qed (auto simp:entails_def model_def)

end